/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/



//This is the main function of the simulator.
//It creates the modules, links them together and starts the simulation.
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <rng.h>
#include <trgen.h>
#include <trdump.h>
#include <receiver.h>
#include <intransit.h>
#include <state.h>
#include <codeset.h>
#include <txsched.h>
#include <stateest.h>
#include <transmitter.h>
#include <statshandler.h>

#include <debug.h>


//The modules
Scheduler *sched;
TrGen *trgen;
TrDump *trdump;
InTransit *it;
Codeset *codeset;
Receiver *rcvr;
TxSched *txsched;
StateEstimator *stateest;
Transmitter *trn;
StatsHandler * stats;

//Input traffic file and output traffic file
char inputfile[20];
char outputfile[20];

//Propagation speed of signal in optical fiber
int propSpeed = (int)2E8;

//When system is in steady state, this is set to 1
int steadystate =0;

//Turn debugging for the main functions ON/OFF (0=OFF, 1=ON)
int debug = 0;


//Dump the input parameters to the debug file, for cross checking.
void DumpParams();

//Check point settings
enum {CP_NO=0, CP_READ, CP_WRITE, CP_RW};

int main(int argc,  char **argv)
{

	if(argc != 31)
	{
		printf("Usage: simone <seed> <N> <A> <w> <K> <Modulation> <Number of codewords> <Chip rate in c/s> <Interarrival distribution> <Packetsize distribution>  <Input traffic file> <Output traffic file> <Average packet size in bytes> <Variance of packet size> <Normalized offered load> <Number of packets> <Topology> <Number of nodes> <Topology distribution>  <Average distance of node from coupler in m> <Tx scheduling algorithm> <Threshold> <Defer alg> <State estimation algorithm> <State estimation samples (n_e) (bits)> <Processing delay (bits)> <Sampling> <Sampling frequency> <Initial state check point> <Steady state packets> \n");
		printf("<Modulation> 0-OOK 1-CCM\n");
		printf("<Topology> 0-Star/Bus\n");
		printf("<Sampling> 0-No 1-Yes\n");
		printf("<Transmission scheduling algorithm> 0-Aloha 1-Selfish 2-Threshold 3-Overlap section 4-Overlap size 5-Admission control 6-Distributed perfect scheduling 7-Optimal 8-Minimimum cross correlation.\n");
		printf("<Defer algorithm> 0-None 1-Binary Exponential Backoff\n");
		printf("<State estimation algorithm> 0-Perfect 1-On demand 2-Continuous 3-Reduced\n");
		printf("<Distributions> 0-Exponential 1-Uniform 2-Deterministic 3-Trimodal 4-From file 5-Normal 6-Pareto 7-Bimodal\n");
		printf("<Initial state check point> 0-No checkpoint 1-Read 2-Write 3-Read and Write\n");
		printf("<Steady state> Number of packets until steady state is reached\n");
		exit(0);
	}


	int seed = atoi(argv[1]);
	srandom(seed);
	
	int N = atoi(argv[2]);
	int A = atoi(argv[3]);
	int w = atoi(argv[4]);
	int K = atoi(argv[5]);
	int numcodewords = atoi(argv[7]);

	int intarr = atoi(argv[9]);
	int pktsize = atoi(argv[10]);
	strcpy(inputfile, argv[11]);
	strcpy(outputfile, argv[12]);
	
	debug = debugon("main");

	dprintf("main.cc: main(): Arguments: Seed: %d\n", seed);
	dprintf("main.cc: main(): Arguments: Input traffic file: %s\n", inputfile);

	sched = new Scheduler();
	codeset = new Codeset(N, A, w, K, numcodewords);
	codeset->chiprate = atof(argv[8]);
	it = new InTransit(A, N);

	double avgpktsize = atof(argv[13]);
	double varpktsize = atof(argv[14]);
	double offeredload = atof(argv[15]);
	trgen = new TrGen(intarr, pktsize, avgpktsize, offeredload, seed);
	trgen->varpktsize = varpktsize;
	trgen->totalpkts = atoi(argv[16]);
	trgen->numnodes = atof(argv[18]);
	trgen->topologydist = atoi(argv[19]);
	trgen->avgdistance = atof(argv[20]);
	int modulation = atoi(argv[6]);
	trgen->modulation = modulation ;


	int txalg = atoi(argv[21]);
	double thparam = atof(argv[22]);
	switch (txalg)
	{
		case TXSCHED_ALOHA:	
			txsched = new TxSched_Aloha();
			break;
		case TXSCHED_SELFISH:
			txsched = new TxSched_Selfish(N, A, w, K);
			break;
		case TXSCHED_THRESHOLD:
			txsched = new TxSched_Threshold(N, A, w, K, thparam);
			break;
		case TXSCHED_OVERLAPSECTION:
			txsched = new TxSched_OverlapSection(N, A, w, K, thparam);
			break;
		case TXSCHED_OVERLAPSIZE:
			txsched = new TxSched_OverlapSize(N, A, w, K, thparam);
			break;
		case TXSCHED_ADMISSIONCONTROL:
			txsched = new TxSched_AdmissionControl(N, A, w, K, thparam);
			break;
		case TXSCHED_DPS:
			txsched = new TxSched_DPS(N, A, w, K, thparam);
			break;
		case TXSCHED_OPTIMALSTATE:
			txsched = new TxSched_OptimalState(N, A, w, K);
			break;
		case TXSCHED_MINCROSSCORR:
			txsched = new TxSched_MinCrossCorr(N, A, w, K, thparam);
			break;
	}

	int defer = atoi(argv[23]);	
	int stalg = atoi(argv[24]);
	int window = atoi(argv[25]);
	stateest = new StateEstimator(N, A);		
	stateest->alg = stalg;
	stateest->ne = window;
	stateest->topology = atoi(argv[17]);
	stateest->sampling = atof(argv[27]);
	stateest->samplingfreq = atof(argv[28]);

	int delay = atoi(argv[26]);
	trn = new Transmitter();
	txsched->deferalg = defer;
	txsched->pdelay = delay;
	trdump = new TrDump();
	stats = new StatsHandler( N, A, 0.000001, 1000);    //Default stats collection interval = 1us.
	stats->normoffload = offeredload;

	trgen->steadystatepktcnt = atoi(argv[30]);	
	
	if(modulation ==OOK)	
		rcvr = new CorrelationReceiver(N, A, w, K);
	else if (modulation == CCM)
		rcvr = new CCMReceiver(N, A, w, K);

	DumpParams();
	
	dprintf("main.cc: main(): Initialization complete. Reading traffic\n");


	int cp = atoi(argv[29]);
	if((cp == CP_READ) || (cp ==CP_RW))
	{
		double t =0;
		int p =0;
		it->readcheckpoint(&t, &p);
		trgen->id = p;
		trgen->lastarrival = t;
		sched->currenttime = t;
		int *state = stateest->calcstate(0);
		dprintf("main.cc: main(): Initialization complete. True state:\n");
		stateest->dumpstate(state);
		int chips = 0;
		for (int count1 =0; count1 < A; count1++)
			for (int count =N-1; count >=0; count--)
				chips+=state[count1*N + count];
		int numonline =  chips/w;
		printf("main.cc: main(): Number on line: %d\n", numonline); 		
	}
	
	trgen->read();

	while (1)
		if(!(sched->run()))
			break;

	dprintf("main.cc: main(): No more events to schedule. Exiting\n");
	if((cp == CP_WRITE) || (cp == CP_RW))
	{
		it->dumpcheckpoint();
	}

	trdump->dumpstats();
	stateest->dumpstats();
	txsched->dumpstats();
	stats->dumpstats();
	
	delete trdump;
	delete it;
	delete trgen;
	delete sched;
	exit(0);
}

void DumpParams()
{

	
	dprintf("main.cc: DumpParams(): N: %d\n", codeset->N);
	dprintf("main.cc: DumpParams(): A: %d\n", codeset->A);
	dprintf("main.cc: DumpParams(): w: %d\n", codeset->w);
	dprintf("main.cc: DumpParams(): K: %d\n", codeset->K);
	dprintf("main.cc: DumpParams(): B: %.6f\n", codeset->chiprate);
	dprintf("main.cc: DumpParams(): Numcodewords: %d\n", codeset->numcodewords);
	dprintf("main.cc: DumpParams(): Arrival dist %d\n", trgen->intarrdist);
	dprintf("main.cc: DumpParams(): Packet size dist %d\n", trgen->pktsizedist);
	dprintf("main.cc: DumpParams(): Avg pkt size %.6f\n", trgen->avgpktsize);
	dprintf("main.cc: DumpParams(): Offered load: %.6f\n", trgen->offeredload);
	dprintf("main.cc: DumpParams(): Num packets: %d\n", trgen->totalpkts);
	dprintf("main.cc: DumpParams(): Topology %d\n", stateest->topology);
	dprintf("main.cc: DumpParams(): Numnodes %.6f\n", trgen->numnodes);
	dprintf("main.cc: DumpParams(): Average distance %.6f\n", trgen->avgdistance);
	dprintf("main.cc: DumpParams(): Average distance from bus coupler %.6f\n", trgen->avgdc);
	//dprintf("main.cc: DumpParams(): Tx Scheduling alg: %d\n", txsched->txalg);
	//dprintf("main.cc: DumpParams(): Threshold: %.6f\n", txsched->thparam);
	dprintf("main.cc: DumpParams(): State estimation alg: %d\n", stateest->alg);
	dprintf("main.cc: DumpParams(): Window: %.6f\n", stateest->ne);
	dprintf("main.cc: DumpParams(): Processing delay: %.6f\n", txsched->pdelay);
	dprintf("main.cc: DumpParams(): Sampling: %.6f\n", stateest->sampling);
	dprintf("main.cc: DumpParams(): Sampling frequency : %.6f\n", stateest->samplingfreq);

}
