/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <debug.h>
#include <packet.h>
#include <eventhandler.h>
#include <scheduler.h>
#include <queue.h>
#include <state.h>
#include <codeset.h>
#include <intransit.h>
#include <stateest.h>

//Packet class functions
//Contains functions that allow calculation of the packets position (boundary) at any point in time/space
//and calculation of the this packets' contribution to the state of the line at a point in time/space.

extern Codeset *codeset;
extern int propSpeed;
extern Scheduler * sched;
extern InTransit *it;

extern StateEstimator *stateest;
Packet::Packet(): next(NULL), prev(NULL), id()
{
	
	dprintf("packet.cc: Packet::Packet(): Construction\n"); 		
	status = PACKET_WAITINGFORTX;
	backoffs =0;
	linecodewords = NULL;
	state = NULL;
	data = NULL;
	defers =0;
}


//The packet format supports two forms of modulation:
//1. ON/OFF keying (OOK)
//2. Code cyclic modulation (CCM): P. Kumar, R. Omrani, J. Touch, A. Willner and P. Saghari, "Three fold increase in the number of users in an OCDMA system using novel code cycle modulation", Under submission, 2004


void Packet::initialize()
{
	dprintf("packet.cc: Packet::Packet(): Intialization: %d\n", id); 		
 	datarng = new RNG();
	if(datarng == NULL)
	{
		printf("Error: packet.cc: Memory allocation failed\n");
		exit(-1);
	}
	for(int count =0; count < id; count++)
	{
		datarng->reset_next_substream();
	}
	if(modulation == OOK)
	{
		symsize = bytesize * 8;
		data = new int[symsize];
		if(data == NULL)
		{
			printf("Error: packet.cc: Memory allocation failed\n");
			exit(-1);
		}
	}
	else if (modulation == CCM)
	{
		//Modulation is CCM: Set size equal to log of packet size.
		//Size here is the number os symbols in the packet
		double bits = (8.0* (double) bytesize);
		double scale = log(N)/log(2);
		symsize = (int)(bits/scale);

		dprintf("packet.cc: Packet::initialize(): Bits: %.6f Scale: %.6f\n", bits, scale); 		
		dprintf("packet.cc: Packet::initialize(): Packet size: Bytes: %d Symbols: %d\n", bytesize, symsize); 		
		data = new int[symsize];
		if(data == NULL)
		{
			printf("Error: packet.cc: Memory allocation failed\n");
			exit(-1);
		}
	}

//The data of the packet is genewrated randomly.
//For OOK, the bit is either 0 or 1 with equal probability (0.5)
//For CCM, the symbol can be one of N phase shifts, chosen uniform randomly.
				
	datarng->reset_start_substream();
 	for(int count=0; count <symsize; count++)
	{

		if(modulation == OOK)
		{
			data[count] = (int) (datarng->uniform(2));

		}

		else if(modulation == CCM)
		{
			//Modulation is CCM.
			//Store the phase shift of the codeword
			int shift = datarng->uniform(N);
			data[count] = shift;
			dprintf("packet.cc: Packet::initialize(): Shift: %d\n", shift); 		
		}
	}
	delete datarng;
}


void Packet::checkdata()
{

	
 	for(int count=0; count <symsize; count++)
	{
		if(modulation == OOK)
		{
			if((data[count] != 0 ) &&( data[count] != 1))
				printf("packet.cc: Packet::initialize(): Data generation error: %d, %d\n", id, count); 		
			
		}

	}

}



//This copy constructor is meant for use by the functions in state.cc
//Note that the data is not copied. Only the original packet (which is stored in the intransit object) 
//will have the data. This is intentional - to reduce memory usage.
//If the data is needed and only a copy is avaialble, search in the intransit object 
//for the original packet.
Packet::Packet(Packet & p)
{
	
	next = NULL;
	prev = NULL;
	id = p.id;

	dprintf("packet.cc: Packet::Packet(): Copy: %d\n", id); 		
	srcaddr = p.srcaddr;
	srcport = p.srcport;
	dstaddr = p.dstaddr;
	dstport = p.dstport;
	symsize = p.symsize;
	bytesize = p.bytesize;

	codeword = p.codeword;
	modulation = p.modulation;
	defers = p.defers;
	
	depx = p.depx;

	departuretime = p.departuretime;;
	arrivaltime = p.arrivaltime;;

	chiprate = p.chiprate;
 	bitrate = p.bitrate;
	
	status = p.status;
	A = p.A;
	N = p.N;
	w = p.w;

	backoffs=p.backoffs;


//Don't bother copying the data or state 
	data = NULL;
	state = NULL;
}

Packet::~Packet()
{
	dprintf("packet.cc: Packet::~Packet(): Destruction: %d\n", id); 		
	if(data != NULL)
		delete [] data;
	if(state != NULL)
		delete [] state;

//DPS code
/*	LineCodeword *t = linecodewords;
	while(t != NULL)
	{
		LineCodeword *l = t;
		t = l->next;
		delete l;
	}
*/



}


void Packet::calcboundary(double *head, double *tail)
{


//Given a packet (i.e. size and departure time), this function calculates the boundary (head/tail) of the packet at the 
//current time. 
//The head and tail of the packets are points on the line
//They are measured in terms of the distance from the coupler(headend).

//The coupler is the origin (zero distance)
//Distance on the transmit fiber is +ve while distance on the receive fiber is -ve.

//Topology is a flat bus structure. y = z = 0. x measures the distances from the headend.
		dprintf("packet.cc: Packet::calcboundary(): Checking state for packet: %d\n", id); 		
		dprintf("packet.cc: Packet::calcboundary(): Departure time: %.12f\n", departuretime); 		
		dprintf("packet.cc: Packet::calcboundary(): Size in bytes: %ld\n", bytesize); 		
		dprintf("packet.cc: Packet::calcboundary(): Size in symbols: %ld\n", symsize); 		
		dprintf("packet.cc: Packet::calcboundary(): Bit rate: %.12f\n", bitrate); 		
		
		double transittime = sched->currenttime - departuretime;
		dprintf("packet.cc: Packet::calcboundary(): Transit time: %.12f s\n", transittime); 		
	 	
		//Measured in meters
		double transitdistance = propSpeed * transittime; 
		dprintf("packet.cc: Packet::calcboundary(): Transit distance: %.12f m \n", transitdistance); 		
	

		//Chip length in meters
		double chiplength = (1.0/chiprate) * propSpeed;	
		dprintf("packet.cc: Packet::calcboundary(): Length of chip: %.12f m \n", chiplength); 		
		
		//Neglect any fraction of the transmission time - assume everything is on pico second boundaries
		double packetlength = (double)symsize * codeset->N;
		dprintf("packet.cc: Packet::calcboundary(): Packet length: %.12f chips\n", packetlength); 		
		
		double transmissionspan = packetlength * chiplength;
		dprintf("packet.cc: Packet::calcboundary(): Transmission span: %.12f m\n", transmissionspan); 		

		*head = depx - transitdistance;
		*tail = *head +  transmissionspan;
		dprintf("packet.cc: Packet::calcboundary(): Packet head: %.12f m\n", *head); 		
		dprintf("packet.cc: Packet::calcboundary(): Packet tail: %.12f m\n", *tail); 		
}






int * Packet::calcstate(double x, int allones)
{

//This function calculates the contibution of this packet to the state of the line at 
//(x at the current time.
//If the packet is on the line at point x, then its codeword is on the line at that point. 
//The codeword may be ON or OFF and it may be shifted.
//This function calculates the shifted version of the codeword due to this packet at the point x.
//The state is calculated by determining the data and the codeword at point x.

//Note that that state is measured between x and x+dx
//y and z are not used. Only dimension x is used.
//(x, x+dx) is called the window.


//The topology is a flat bus (star) structure. x measures the distance from the headend.
//Though it may be a star, the simulation simulates it as a bus
//The two can be proved to be equivalent


	int * state = NULL;
	dprintf("packet.cc: Packet::calcstate(): Checking state for packet: %d\n",id); 		

	double packethead;
	double packettail;
	calcboundary(&packethead, &packettail);
	dprintf("packet.cc: Packet::calcstate(): Packet head: %.12f m\n", packethead); 		
	dprintf("packet.cc: Packet::calcstate(): Packet tail: %.12f m\n", packettail); 		


	//Chip length in meters
	double chiplength = (1.0/chiprate) * propSpeed;	
	dprintf("packet.cc: Packet::calcstate(): Length of chip: %.12f m \n", chiplength); 		
	double x1 = x;
	double x2 = x + N*chiplength;

//State is measured from the point x to a point x-dx, where dx corresponds to the length of the codeset (N)
//Each element of the state array represents a chip on the line.
	dprintf("packet.cc: Packet::calcstate(): x1: %.12f m\n", x1); 		
	dprintf("packet.cc: Packet::calcstate(): x2: %.12f m\n", x2); 		

//Convert to number of chips (integer) for compariosn.
	int ph=0;
	int pt=0;
	int ix1 =0;
	int ix2 =0;

	if (packethead >= 0)
		ph = int((packethead/chiplength) + 0.5);
	else
		ph = int((packethead/chiplength) - 0.5);

	if (packettail >= 0)
		pt = int((packettail/chiplength) + 0.5);
	else
		pt = int((packettail/chiplength) - 0.5);


	if (x1 >= 0)
		ix1 = int((x1/chiplength) + 0.5);
	else
		ix1 = int((x1/chiplength) - 0.5);

	if (x2 >= 0)
		ix2 = int((x2/chiplength) + 0.5);
	else
		ix2 = int((x2/chiplength) - 0.5);

//3 cases: 
//1. Packet completly overlaps with the window
//2. Packet tail overlaps with window.
//3. Packet head overlaps with window.
	dprintf("Packethead: %.6f Packettail: %.6f x1: %.6f x2: %.6f\n", packethead, packettail, x1, x2);
	dprintf("Packethead: %d Packettail: %d x1: %d x2: %d\n", ph, pt, ix1, ix2);


	if ((ph <= ix1 ) && (pt  >= ix2)) 
	{

		dprintf("packet.cc: Packet::calcstate(): Found packet in region 1:  %d\n", id); 	
		double offset = x1 - packethead;
		dprintf("packet.cc: Packet::calcstate(): Offset: %.6f\n", offset); 	

		int chipoffset =0;
		if (offset >= 0)
			chipoffset = int((offset/chiplength) + 0.5);
		else
			chipoffset = int((offset/chiplength) - 0.5);

		dprintf("packet.cc: Packet::calcstate(): Chip offset: %d\n", chipoffset); 	
		int bit = chipoffset/N;

		chipoffset = chipoffset % (int) N;	
		if(chipoffset < 0)
			chipoffset = -chipoffset;
	
		dprintf("packet.cc: Packet::calcstate(): Checking bit: %d\n", bit); 	
		dprintf("packet.cc: Packet::calcstate(): Size in bytes: %d\n", bytesize); 	
		dprintf("packet.cc: Packet::calcstate(): Size in symbols %d\n", symsize); 	
		dprintf("packet.cc: Packet::calcstate(): Chip offset: %d\n", chipoffset); 	
		dprintf("packet.cc: Packet::calcstate(): Data: %d\n", data); 	

		//If chipoffset is not equal to zero, then two codewords span the
		//window.

		if (chipoffset == 0)
		{
			if(modulation == OOK) 
			{
				if((data[bit] == 1) || (allones == 1))
				{
					//Return the codeword unchanged.
					dprintf("packet.cc: Packet::calcstate(): Data one\n"); 	
					state = codeset->LeftShiftCodeword(codeword, chipoffset);
					return state;	
				}
				else if(data[bit] == 0)
				{
					dprintf("packet.cc: Packet::calcstate(): Data zero\n"); 	
					state = new int[A*N];
					if(state == NULL)
					{
						printf("Error: packet.cc: Memory allocation failed\n");
						exit(-1);
					}
					for (int count = 0; count < A*N; count++)
						state[count] = 0;
					return state;	
				}
			}
			else if(modulation == CCM)
			{
					dprintf("packet.cc: Packet::calcstate(): CCM\n"); 	
					dprintf("packet.cc: Packet::calcstate(): Phase shift: %d\n", data[bit]); 	
					state = codeset->LeftShiftCodeword(codeword, data[bit]);
					return state;	

			}
		}
		else
		{
			dprintf("packet.cc: Packet::calcstate(): Overlapping window and packet\n"); 	
			if(modulation == OOK)
			{		
				dprintf("packet.cc: Packet::calcstate(): Modulation: OOK\n"); 	
				dprintf("packet.cc: Packet::calcstate(): Bits: %d %d\n", data[bit], data[bit+1]);
		
				//2 codewords span the window: 4 cases: 00 01 10 11
				if(((data[bit] == 1) &&(data[bit+1] == 1)) || (allones == 1))
				{
					dprintf("packet.cc: Packet::calcstate(): Case 11\n"); 	
					state = codeset->LeftShiftCodeword(codeword, chipoffset);
					return state;	
				}
				else if((data[bit] == 0) &&(data[bit+1] == 0))
				{
					dprintf("packet.cc: Packet::calcstate(): Case 00\n"); 	
					state = new int[A*N];
					if(state == NULL)
					{
						printf("Error: packet.cc: Memory allocation failed\n");
						exit(-1);
					}
					for (int count = 0; count < A*N; count++)
						state[count] = 0;
					return state;	
				}
				else if((data[bit] == 1) &&(data[bit+1] == 0))
				{

					dprintf("packet.cc: Packet::calcstate(): Case 10\n"); 	
					state = codeset->LeftShiftCodeword(codeword, chipoffset);
					//Zero the remaining chips
					for (int count = chipoffset; count < A*N; count++)
						state[count] = 0;
					return state;	
				}
				else if((data[bit] == 0) &&(data[bit+1] == 1))
				{
					dprintf("packet.cc: Packet::calcstate(): Case 01\n"); 	

					state = codeset->LeftShiftCodeword(codeword, chipoffset);
					//Zero the remaining chips
					for (int count = 0; count < chipoffset; count++)
						state[count] = 0;
					return state;	
				}
				dprintf("packet.cc: Packet::calcstate(): Serious error, data corruption\n"); 	

			}
			else if(modulation == CCM)
			{
				
					dprintf("packet.cc: Packet::calcstate(): Case CCM\n"); 	
					state = codeset->LeftShiftCodeword(codeword, data[bit]);
					int * state2 = codeset->LeftShiftCodeword(codeword, data[bit+1]);
					dprintf("packet.cc: Packet::calcstate(): Case CCM\n"); 	
					//Zero the remaining chips
					for (int count = chipoffset; count < A*N; count++)
						state[count] = state2[count];
					delete state2;
					return state;	

			}
		}
 	
		dprintf("packet.cc:Packet::calcstate(): Code: \n");
		for (int count1 =0; count1 < A; count1++)
			for (int count =N-1; count >=0; count--)
		{
				dprintf("%d ", codeset->set[codeword*(A*N) + N*count1  + count]);
		}
		::dprintf("\n");	
		
		dprintf("packet.cc: Packet::calcstate(): Offset in front of head: %d (chips)\n", chipoffset); 		
		dprintf("packet.cc: Packet::calcstate(): Packet shifted\n"); 		
		dprintf("packet.cc:Packet::calcstate(): Shifted Codeword: \n");
		for (int count1 =0; count1 < A; count1++)
			for (int count =N-1; count >=0; count--)
			{
				::dprintf("%d ", state[count1*N + count]);	
			}
		::dprintf("\n");	

	
	}
	else if ((ph  >= ix1 ) && (ph  < ix2) && (pt   >= ix2)) 
	{

		dprintf("packet.cc: Packet::calcstate(): Found packet in regioni 2:  %d\n", id); 	
		double offset = x1 - packethead;
		dprintf("packet.cc: Packet::calcstate(): Offset: %.6f\n", offset); 	

		int chipoffset =0;
		if (offset >= 0)
			chipoffset = int((offset/chiplength) + 0.5);
		else
			chipoffset = int((offset/chiplength) - 0.5);

		int bit = chipoffset/N;
		chipoffset = chipoffset % (int) N;	
		if(chipoffset < 0)
			chipoffset = -chipoffset;
		dprintf("packet.cc: Packet::calcstate(): Chip Offset: %d\n", chipoffset); 	
		if(modulation == OOK)
		{
			if((data[bit] == 1) || (allones == 1))
			{
				state = codeset->LeftShiftCodeword(codeword, N - chipoffset);
				//Zero the remaining chips
				for (int count = N-chipoffset; count < N; count++)
					state[count] = 0;
			}
			else
			{
				state = new int[A*N];
				if(state == NULL)
				{
					printf("Error: packet.cc: Memory allocation failed\n");
					exit(-1);
				}
				for (int count = 0; count < A*N; count++)
					state[count] = 0;
			}
		}
		else if(modulation == CCM)
		{

				state = codeset->LeftShiftCodeword(codeword, data[bit]);
				//Zero the remaining chips
				for (int count = N-chipoffset; count < N; count++)
					state[count] = 0;
		}

	
		dprintf("packet.cc: Packet::calcstate(): Offset in front of head: %d (chips)\n", chipoffset); 		
		dprintf("packet.cc: Packet::calcstate(): Packet shifted\n"); 		
		dprintf("packet.cc:Packet::calcstate(): Shifted Codeword: \n");
		for (int count1 =0; count1 < A; count1++)
			for (int count =N-1; count >=0; count--)
			{
				::dprintf("%d ", state[count1*N + count]);	
			}
		::dprintf("\n");	
		return state;	



	}
	else if ((ph < ix1 ) && (pt <= ix2) && (pt  > x1))
	{
		dprintf("packet.cc: Packet::calcstate(): Found packet in region 3:  %d\n", id); 	
		double offset = x1 - packethead;
		dprintf("packet.cc: Packet::calcstate(): Offset: %.6f\n", offset); 	

		int chipoffset =0;
		if (offset >= 0)
			chipoffset = int((offset/chiplength) + 0.5);
		else
			chipoffset = int((offset/chiplength) - 0.5);

		int bit = chipoffset/N;
		chipoffset = chipoffset % (int) N;	
		if(chipoffset < 0)
			chipoffset = -chipoffset;

		if(modulation == OOK)
		{
			if((data[bit] == 1) || (allones == 1))
			{
				state = codeset->LeftShiftCodeword(codeword, chipoffset);
				//Zero the remaining chips
				for (int count = chipoffset; count < A*N; count++)
					state[count] = 0;
			}
			else
			{
				state = new int[A*N];
				if(state == NULL)
				{
					printf("Error: packet.cc: Memory allocation failed\n");
					exit(-1);
				}
				for (int count = 0; count < A*N; count++)
					state[count] = 0;
			}
		}
		else if(modulation == CCM)
		{
				state = codeset->LeftShiftCodeword(codeword, data[bit]);
				//Zero the remaining chips
				for (int count = chipoffset; count < A*N; count++)
					state[count] = 0;


		}
		dprintf("packet.cc: Packet::calcstate(): Offset in front of head: %d (chips)\n", chipoffset); 		
		dprintf("packet.cc: Packet::calcstate(): Packet shifted\n"); 		
		dprintf("packet.cc:Packet::calcstate(): Shifted Codeword: \n");
		for (int count1 =0; count1 < A; count1++)
			for (int count =N-1; count >=0; count--)
			{
				::dprintf("%d ", state[count1*N + count]);	
			}
		::dprintf("\n");	

		return state;	

	}
	else
	{
		dprintf("packet.cc: Packet::calcstate(): Packet not in region:  %d\n", id); 		
		state = new int[A*N];
		if(state == NULL)
		{
			printf("Error: packet.cc: Memory allocation failed\n");
			exit(-1);
		}
		for (int count = 0; count < A*N; count++)
			state[count] = 0;
		return state;	
	}
	return state;	

}




