/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <receiver.h>
#include <trdump.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>


extern Scheduler * sched;
extern InTransit * it;
extern char outputfile[20];
extern Codeset *codeset;

extern int propSpeed;


//Implements the functions of a receiver. 
//Give a packet and the corresponding state at a point in time/space, it calculates whether that packet will be received correctly
//Currently only the hard-limiting correlation receiver is implemented.
//Once a packet's tail crosses the coupler (head-end), the receiver is called to 
//check if it has been received correctly.
//The packet is broken into sections and each section is examined for errors. 
//The receiver operates on a section and the receiving codeword


Receiver::Receiver():EventHandler()
{
	debug = debugon("receiver");
}

Receiver::~Receiver()
{
}


void Receiver::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend


}
//The event handling function of the Receiver. It dequeues a packet and
//sends it to the next component.
void Receiver::handleevent(Packet *p)
{
	dtprintf("receiver.cc: Receiver::handleevent(): Entered function\n");
	

	//Pass it on to the next component in the path.
	dtprintf("receiver.cc: Receiver::handleevent(): Packet id: %d\n", p->id);
	dtprintf("receiver.cc: Receiver::handleevent(): Packet processing complete, passing it on to the next component.\n");


}




void Receiver::dumpstats()
{
	printf("receiver.cc: Receiver::dumpstats():\n");
}

int Receiver::receive(Section * s, Packet *p)
{

	return 0;
}

CorrelationReceiver::~CorrelationReceiver()
{
}

CorrelationReceiver::CorrelationReceiver(int n, int a, int W, int K): Receiver()
{
	A = a;	
	N = n;
	w = W;
}



int CorrelationReceiver::receive(Section * s, Packet *rp)
{
	double x = s->head;

	dtprintf("receive.cc: Receive::receive(): Receiving packet: %d\n", rp->id); 		
	dtprintf("receive.cc: Receive::receive(): Receiving codeword: %d\n", rp->codeword); 		
	dtprintf("receive.cc: Receive::receive(): Calculating state\n"); 		
	dtprintf("receive.cc: Receive::receive(): Calculating state at : %.6f, %.6f, %.6f\n", x, 0, 0); 		
	
	double chiptime = (1.0/codeset->chiprate);
	double chiplength = chiptime * propSpeed;
	int chips = (int)((s->tail - s->head)/chiplength);
	int bits = chips/codeset->N;

	double packethead;
	double packettail;
	rp->calcboundary(&packethead, &packettail);
	dtprintf("receive.cc: Receive::receive(): Packet head: %.6f\n", packethead); 		
	dtprintf("receive.cc: Receive::receive(): Packet tail: %.6f\n", packettail); 		
	dtprintf("receive.cc: Receive::receive(): Section head: %.6f\n", s->head); 		
	dtprintf("receive.cc: Receive::receive(): Section tail: %.6f\n", s->tail); 		
	dtprintf("receive.cc: Receive::receive(): Chips: %d\n", chips); 		
	dtprintf("receive.cc: Receive::receive(): Bits: %d\n", bits); 		


//Condition for correct reception
//Refer to [Purushotham Kamath, "Interference Avoidance: Media Access Control for Optical CDMA", PhD. dissertation]  for details. 

//First check if the packet will survive if all other codewords are ON
//If it does, then it will survive. 
	checkforerror(s, rp, 1, 0);
	if(rp->status == PACKET_OK)
	{
		return 0;
	}



//Condition for a packet error
//Refer to [Purushotham Kamath, "Interference Avoidance: Media Access Control for Optical CDMA", PhD. dissertation]  for details. 
//If not, check very single bit for errors. It may still survive if 
//the other codwords are not switched ON in the right combinations
//For every byte in the section
/*	for(int counta = 0; counta < bits; counta++)
	{
		checkforerror(s, rp, 0, counta);
	}
	return 0;
*/
	return 0;
}




void CorrelationReceiver::checkforerror(Section * s, Packet *rp, int allones, int counta)
{
	double x = s->head;
	double chiptime = (1.0/codeset->chiprate);
	double chiplength = chiptime * propSpeed;

	int * rpstate = rp->calcstate(x + counta*N*chiplength, allones); 

	//Note it may be shifted because it is relative to the section head.
	dtprintf("receive.cc: Receive::receive(): Checking state at %.6f: \n", x+counta*N*chiplength); 		
	dtprintf("receive.cc: Receive::receive(): Packet id: %d\n", rp->id); 		
	dtprintf("receive.cc: Receive::receive(): Codeword: \n"); 		
	for (int count1 =0; count1 < rp->A; count1++)
	{
		for (int count =rp->N -1; count >=0; count--)
		{
			dprintf("%d ", rpstate[count1*N + count]);
		
		/*	if(rpstate[count1*N + count] == 1)
				printf("%d ", count);
		*/

		}	
		dprintf("\n");
	}


	int * state = new int [(rp->A) *(rp->N)];
	for (int count = 0; count < (rp->A)*(rp->N); count++)
		state[count] =  0;

	int pcount =0; 

	Packet*p = s->q->head;
	while(p!= NULL)
	{


		//Only the intransit object contains the actual data
		//So use it to determine state.
		Packet *datap = it->q->head;
		while(datap != NULL )
		{
			if(datap->id == p->id)
				break;
			datap=datap->next;
		}	

		dtprintf("receive.cc: Receive::receive(): Checking packet: %d\n", p->id); 		
		int * pstate = datap->calcstate(x+counta*N*chiplength, allones);
		for (int count = 0; count < rp->A*rp->N; count++)
		{
			state[count]+= pstate[count];

		}	
		delete pstate;	
		pcount++;
		p = p->next;
	}

	dtprintf("receive.cc: Receive::receive(): Numonline: %d\n", pcount); 		
	dtprintf("receive.cc: Receive::receive(): State: \n"); 		
	for (int count1 =0; count1 < rp->A; count1++)
	{
		for (int count =rp->N -1; count >=0; count--)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}
	dprintf("\n"); 		
	


	dprintf("receive.cc: Receive::receive(): Checked all packets\n"); 		
	int overlaps = 0;	
	for (int count = 0; count < rp->A*rp->N; count++)
		if ((rpstate[count] == 1) && (state[count] > 1))
		{
			overlaps++;
		}
	

/*
	for (int count = 0; count < rp->A*rp->N; count++)
		if ((rpstate[count] == 1))
			printf("%d ", state[count]);
	printf("\n"); 		
*/


	//printf("receive.cc: Receive::receive(): Overlaps: %d\n", overlaps); 		

//Condition for correct reception.
	if(overlaps == rp->w)
	{
		dtprintf("receive.cc: Receive::receive(): Error: %d\n", overlaps); 		
		rp->status = PACKET_ERROR_MUI;
	}
	else
	{
		dtprintf("receive.cc: Receive::receive(): No error: %d\n", overlaps); 		
	}
	delete state;
	delete rpstate;
}



void CorrelationReceiver::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend


}
//The event handling function of the CorrelationReceiver. It dequeues a packet and
//sends it to the next component.
void CorrelationReceiver::handleevent(Packet *p)
{
	dtprintf("receiver.cc: CorrelationReceiver::handleevent(): Entered function\n");
	

	//Pass it on to the next component in the path.
	dtprintf("receiver.cc: CorrelationReceiver::handleevent(): Packet id: %d\n", p->id);
	dtprintf("receiver.cc: CorrelationReceiver::handleevent(): Packet processing complete, passing it on to the next component.\n");


}

void CorrelationReceiver::dumpstats()
{
	printf("receiver.cc: CorrelationReceiver::dumpstats():\n");
}



CCMReceiver::~CCMReceiver()
{
}

CCMReceiver::CCMReceiver(int n, int a, int W, int K): Receiver()
{
	A = a;	
	N = n;
	w = W;
}


int CCMReceiver::receive(Section * s, Packet *rp)
{
	double x = s->head;

	dtprintf("receive.cc: Receive::receive(): Receiving packet: %d\n", rp->id); 		
	dtprintf("receive.cc: Receive::receive(): Receiving codeword: %d\n", rp->codeword); 		
	dtprintf("receive.cc: Receive::receive(): Calculating state\n"); 		
	dtprintf("receive.cc: Receive::receive(): Calculating state at : %.6f, %.6f, %.6f\n", x, 0, 0); 		
	
	double chiptime = (1.0/codeset->chiprate);
	double chiplength = chiptime * propSpeed;
	int chips = (int)((s->tail - s->head)/chiplength);
	int syms = (chips/codeset->N);

	double packethead;
	double packettail;
	rp->calcboundary(&packethead, &packettail);
	dtprintf("receive.cc: Receive::receive(): Packet head: %.6f\n", packethead); 		
	dtprintf("receive.cc: Receive::receive(): Packet tail: %.6f\n", packettail); 		
	dtprintf("receive.cc: Receive::receive(): Section head: %.6f\n", s->head); 		
	dtprintf("receive.cc: Receive::receive(): Section tail: %.6f\n", s->tail); 		
	dtprintf("receive.cc: Receive::receive(): Chips: %d\n", chips); 		
	dtprintf("receive.cc: Receive::receive(): Symbols: %d\n", syms); 		

//Condition for correct reception

		for(int counta = 0; counta < syms; counta++)
		{
			checkforerror(s, rp, 0, counta);
			if(rp->status == PACKET_ERROR_MUI)
				return -1; 
		}
	//printf("receive.cc: Receive::receive(): No error in packet.\n"); 		



	return 0;
}




void CCMReceiver::checkforerror(Section * s, Packet *rp, int allones, int counta)
{
	double x = s->head;
	double chiptime = (1.0/codeset->chiprate);
	double chiplength = chiptime * propSpeed;

	double packethead;
	double packettail;
	rp->calcboundary(&packethead, &packettail);
	dtprintf("receive.cc: Receive::receive(): Packet head: %.6f\n", packethead); 		
	dtprintf("receive.cc: Receive::receive(): Packet tail: %.6f\n", packettail); 		


	//Find the beginning of the first symbol after the the section start
	double offset = x - packethead;
	dprintf("packet.cc: Packet::calcstate(): Offset: %.6f\n", offset); 	

	int chipoffset =0;
	if (offset >= 0)
		chipoffset = int((offset/chiplength) + 0.5);
	else
		chipoffset = int((offset/chiplength) - 0.5);

	dprintf("packet.cc: Packet::calcstate(): Chip offset: %d\n", chipoffset); 	
	int bit = chipoffset/N;
	chipoffset = chipoffset % (int) N;	
	if(chipoffset < 0)
		chipoffset = -chipoffset;

	int bitoffset = bit + counta; 
	dprintf("receive.cc: Receive::receive(): Bit offset: %d \n", bitoffset);

	//Note it may be shifted because it is relative to the section head.
	dtprintf("receive.cc: Receive::receive(): Checking state at %.6f: \n", packethead + bitoffset*N*chiplength); 		
	int * rpstate = rp->calcstate(packethead + bitoffset*N*chiplength,  allones); 

	dtprintf("receive.cc: Receive::receive(): Packet id: %d\n", rp->id); 		
	dtprintf("receive.cc: Receive::receive(): Codeword: \n"); 		
	for (int count1 =0; count1 < rp->A; count1++)
	{
		for (int count =rp->N -1; count >=0; count--)
		{
			dprintf("%d ", rpstate[count1*N + count]);
		
		/*	if(rpstate[count1*N + count] == 1)
				printf("%d ", count);
		*/

		}	
		dprintf("\n");
	}


	int * state = new int [(rp->A) *(rp->N)];
	for (int count = 0; count < (rp->A)*(rp->N); count++)
		state[count] =  0;

	int pcount =0; 

	Packet*p = s->q->head;
	while(p!= NULL)
	{
		//Only the intransit object contains the actual data
		//So use it to determine state.


		Packet *datap = it->q->head;
		while(datap != NULL )
		{
			if(datap->id == p->id)
				break;
			datap=datap->next;
		}


	
		dtprintf("receive.cc: Receive::receive(): Checking packet: %d\n", p->id); 		
		int * pstate = datap->calcstate(packethead+ bitoffset *N*chiplength, allones);
		dtprintf("receive.cc: Receive::receive(): State: \n"); 		
		for (int count1 =0; count1 < rp->A; count1++)
		{
			for (int count =rp->N -1; count >=0; count--)
			{
				dprintf("%d ", pstate[count1*N + count]);
			}
			dprintf("\n");
		}
		for (int count = 0; count < rp->A*rp->N; count++)
		{
			state[count]+= pstate[count];
		}	
		delete pstate;	
		pcount++;
		p = p->next;
	}

	dtprintf("receive.cc: Receive::receive(): Numonline: %d\n", pcount); 		
	dtprintf("receive.cc: Receive::receive(): State: \n"); 		
	for (int count1 =0; count1 < rp->A; count1++)
	{
		for (int count =rp->N -1; count >=0; count--)
		{
			dprintf("%d ", state[count1*N + count]);
		}
		dprintf("\n");
	}
	dprintf("\n"); 		
	

	int found = 0;

	for (int count = 0; count < rp->N; count++)
	{
		int * shifted = codeset->LeftShiftCodeword(rp->codeword, count);
		dprintf("receive.cc: Receive::receive(): Shifted codeword: \n"); 		
		for (int count1 =0; count1 < rp->A; count1++)
		{
			for (int count2 =rp->N -1; count2 >=0; count2--)
			{
				dprintf("%d ", shifted[count1*N + count2]);
			}	
		}
		dprintf("\n");


		int overlaps = 0;	
		for (int count2 = 0; count2 < rp->A*rp->N; count2++)
		{
			if ((state[count2] >= 1) && (shifted[count2] == 1))
			{
				overlaps++;
			}
		}
		dtprintf("receive.cc: Shift: %d\n", count);
		dtprintf("receive.cc: Data shift: %d\n", rp->data[bitoffset]);
		dtprintf("receive.cc: Overlaps: %d\n", overlaps);
		if((overlaps == rp->w) && (rp->data[bitoffset] != count))
		{
			dtprintf("receive.cc: Error: Error: False positive\n");
			rp->status = PACKET_ERROR_MUI;
			found = 1;
			delete shifted;
			break;
		}
		if((overlaps == rp->w) && (rp->data[bitoffset] == count))
		{
			//It does not matter if there are overlaps in the correct phase shift.
			//Assumption: Receiver will detect correctly if the first shift detected 
			//happens to be right
			dtprintf("receive.cc: Receive::receive(): No error: %d\n", overlaps); 		
			dtprintf("receive.cc: Receive::receive(): Shift: %d\n", count);
			rp->status = PACKET_OK;
			found = 1;
			delete shifted;
			break;
		}
		delete shifted;

	}	

	//Check
	if(found  != 1)
	{
		printf("receive.cc: Receive::receive(): Serious error: Codeword not found\n");
	}
	 

	delete state;
	delete rpstate;
}



void CCMReceiver::process(Packet *p)
{

	//Schedule the event for whenever it passes the headend


}
//The event handling function of the CCMReceiver. It dequeues a packet and
//sends it to the next component.
void CCMReceiver::handleevent(Packet *p)
{
	dtprintf("receiver.cc: CCMReceiver::handleevent(): Entered function\n");
	

	//Pass it on to the next component in the path.
	dtprintf("receiver.cc: CCMReceiver::handleevent(): Packet id: %d\n", p->id);
	dtprintf("receiver.cc: CCMReceiver::handleevent(): Packet processing complete, passing it on to the next component.\n");


}

void CCMReceiver::dumpstats()
{
	printf("receiver.cc: CCMReceiver::dumpstats():\n");
}
